import logging
import os
import platform
import re
import sys
import time
from io import StringIO

import grpc
import numpy as np
import pandas as pd
import pkg_resources
import psutil
import tqdm

import giskard
from giskard.ml_worker.core.giskard_dataset import GiskardDataset
from giskard.ml_worker.core.model_explanation import (
    explain,
    explain_text,
)
from giskard.ml_worker.exceptions.IllegalArgumentError import IllegalArgumentError
from giskard.ml_worker.exceptions.giskard_exception import GiskardException
from giskard.ml_worker.generated.ml_worker_pb2 import (
    DataFrame,
    DataRow,
    EchoMsg,
    ExplainRequest,
    ExplainResponse,
    ExplainTextRequest,
    ExplainTextResponse,
    MLWorkerInfo,
    MLWorkerInfoRequest,
    PlatformInfo,
    RunModelForDataFrameRequest,
    RunModelForDataFrameResponse,
    RunModelRequest,
    RunModelResponse,
    RunTestRequest,
    TestResultMessage,
    UploadStatus,
    FileUploadMetadata,
    FileType,
    StatusCode,
    FilterDatasetResponse,
)
from giskard.ml_worker.generated.ml_worker_pb2_grpc import MLWorkerServicer
from giskard.ml_worker.utils.grpc_mapper import deserialize_dataset, deserialize_model
from giskard.ml_worker.utils.logging import Timer
from giskard.path_utils import model_path, dataset_path

logger = logging.getLogger(__name__)


def file_already_exists(meta: FileUploadMetadata):
    if meta.file_type == FileType.MODEL:
        path = model_path(meta.project_key, meta.name)
    elif meta.file_type == FileType.DATASET:
        path = dataset_path(meta.project_key, meta.name)
    else:
        raise ValueError(f"Illegal file type: {meta.file_type}")
    return path.exists(), path


class MLWorkerServiceImpl(MLWorkerServicer):
    def __init__(self, port=None, remote=None) -> None:
        super().__init__()
        self.port = port
        self.remote = remote

    def echo(self, request, context):
        logger.debug(f"echo: {request.msg}")
        return EchoMsg(msg=request.msg)

    def upload(self, request_iterator, context: grpc.ServicerContext):
        meta = None
        path = None
        progress = None
        for upload_msg in request_iterator:
            if upload_msg.HasField("metadata"):
                meta = upload_msg.metadata
                file_exists, path = file_already_exists(meta)
                if not file_exists:
                    progress = tqdm.tqdm(
                        desc=f"Receiving {upload_msg.metadata.name}",
                        unit="B",
                        unit_scale=True,
                        unit_divisor=1024,
                    )
                    yield UploadStatus(code=StatusCode.CacheMiss)
                else:
                    logger.info(f"File already exists: {path}")
                    break
            elif upload_msg.HasField("chunk"):
                try:
                    path.parent.mkdir(exist_ok=True, parents=True)
                    with open(path, "ab") as f:
                        f.write(upload_msg.chunk.content)
                    progress.update(len(upload_msg.chunk.content))
                except Exception as e:
                    if progress is not None:
                        progress.close()
                    logger.exception(f"Failed to upload file {meta.name}", e)
                    yield UploadStatus(code=StatusCode.Failed)

        if progress is not None:
            progress.close()
        yield UploadStatus(code=StatusCode.Ok)

    def getInfo(self, request: MLWorkerInfoRequest, context):
        logger.info("Collecting ML Worker info")
        installed_packages = (
            {p.project_name: p.version for p in pkg_resources.working_set}
            if request.list_packages
            else None
        )
        current_process = psutil.Process(os.getpid())
        return MLWorkerInfo(
            platform=PlatformInfo(
                machine=platform.uname().machine,
                node=platform.uname().node,
                processor=platform.uname().processor,
                release=platform.uname().release,
                system=platform.uname().system,
                version=platform.uname().version,
            ),
            giskard_client_version=giskard.__version__,
            pid=os.getpid(),
            process_start_time=int(current_process.create_time()),
            interpreter=sys.executable,
            interpreter_version=platform.python_version(),
            installed_packages=installed_packages,
            internal_grpc_port=self.port,
            is_remote=self.remote,
        )

    def runTest(self, request: RunTestRequest, context: grpc.ServicerContext) -> TestResultMessage:
        from giskard.ml_worker.testing.functions import GiskardTestFunctions

        model = deserialize_model(request.model)

        tests = GiskardTestFunctions()
        _globals = {"model": model, "tests": tests}
        if request.reference_ds.file_name:
            _globals["reference_ds"] = deserialize_dataset(request.reference_ds)
        if request.actual_ds.file_name:
            _globals["actual_ds"] = deserialize_dataset(request.actual_ds)
        try:
            timer = Timer()
            exec(request.code, _globals)
            timer.stop(f"Test {tests.tests_results[0].name}")
        except NameError as e:
            missing_name = re.findall(r"name '(\w+)' is not defined", str(e))[0]
            if missing_name == "reference_ds":
                raise IllegalArgumentError("Reference Dataset is not specified")
            if missing_name == "actual_ds":
                raise IllegalArgumentError("Actual Dataset is not specified")
            raise e

        return TestResultMessage(results=tests.tests_results)

    def explain(self, request: ExplainRequest, context) -> ExplainResponse:
        model = deserialize_model(request.model)
        dataset = deserialize_dataset(request.dataset)
        explanations = explain(model, dataset, request.columns)

        return ExplainResponse(
            explanations={
                k: ExplainResponse.Explanation(per_feature=v)
                for k, v in explanations["explanations"].items()
            }
        )

    def explainText(self, request: ExplainTextRequest, context) -> ExplainTextResponse:
        n_samples = 500 if request.n_samples <= 0 else request.n_samples
        model = deserialize_model(request.model)
        text_column = request.feature_name

        if request.feature_types[text_column] != "text":
            raise ValueError(f"Column {text_column} is not of type text")
        text_document = request.columns[text_column]
        input_df = pd.DataFrame({k: [v] for k, v in request.columns.items()})
        if model.feature_names:
            input_df = input_df[model.feature_names]
        (list_words, list_weights) = explain_text(
            model, input_df, text_column, text_document, n_samples
        )
        map_features_weight = dict(zip(model.classification_labels, list_weights))
        return ExplainTextResponse(
            weights={
                k: ExplainTextResponse.WeightsPerFeature(
                    weights=[weight for weight in map_features_weight[k]]
                )
                for k in map_features_weight
            },
            words=list_words,
        )

    def runModelForDataFrame(self, request: RunModelForDataFrameRequest, context):
        model = deserialize_model(request.model)
        ds = GiskardDataset(
            pd.DataFrame([r.columns for r in request.dataframe.rows]),
            target=request.target,
            feature_types=request.feature_types,
            column_types=request.column_types,
        )
        predictions = model.run_predict(ds)
        if model.model_type == "classification":
            return RunModelForDataFrameResponse(
                all_predictions=self.pandas_df_to_proto_df(predictions.all_predictions),
                prediction=predictions.prediction.astype(str),
            )
        else:
            return RunModelForDataFrameResponse(
                prediction=predictions.prediction.astype(str), raw_prediction=predictions.prediction
            )

    def runModel(self, request: RunModelRequest, context) -> RunModelResponse:
        try:
            model = deserialize_model(request.model)
            dataset = deserialize_dataset(request.dataset)
        except ValueError as e:
            if "unsupported pickle protocol" in str(e):
                raise ValueError(
                    "Unable to unpickle object, "
                    "Make sure that Python version of client code is the same as the Python version in ML Worker."
                    "To change Python version, please refer to https://docs.giskard.ai/start/guides/configuration"
                    f"\nOriginal Error: {e}"
                ) from e
            raise e
        except ModuleNotFoundError as e:
            raise GiskardException(
                f"Failed to import '{e.name}'. "
                f"Make sure it's installed in the ML Worker environment."
                "To have more information on ML Worker, please see: https://docs.giskard.ai/start/guides/installation/ml-worker"
            ) from e
        prediction_results = model.run_predict(dataset)

        if model.model_type == "classification":
            results = prediction_results.all_predictions
            labels = {k: v for k, v in enumerate(model.classification_labels)}
            label_serie = dataset.df[dataset.target] if dataset.target else None
            if len(model.classification_labels) > 2 or model.classification_threshold is None:
                preds_serie = prediction_results.all_predictions.idxmax(axis="columns")
                sorted_predictions = np.sort(prediction_results.all_predictions.values)
                abs_diff = pd.Series(
                    sorted_predictions[:, -1] - sorted_predictions[:, -2], name="absDiff"
                )
            else:
                diff = (
                        prediction_results.all_predictions.iloc[:, 1] - model.classification_threshold
                )
                preds_serie = (diff >= 0).astype(int).map(labels).rename("predictions")
                abs_diff = pd.Series(diff.abs(), name="absDiff")
            calculated = pd.concat([preds_serie, label_serie, abs_diff], axis=1)
        else:
            results = pd.Series(prediction_results.prediction)
            preds_serie = results
            if dataset.target and dataset.target in dataset.df.columns:
                target_serie = dataset.df[dataset.target]
                diff = preds_serie - target_serie
                diff_percent = pd.Series(diff / target_serie, name="diffPercent")
                abs_diff = pd.Series(diff.abs(), name="absDiff")
                abs_diff_percent = pd.Series(abs_diff / target_serie, name="absDiffPercent")
                calculated = pd.concat(
                    [preds_serie, target_serie, abs_diff, abs_diff_percent, diff_percent], axis=1
                )
            else:
                calculated = pd.concat([preds_serie], axis=1)

        return RunModelResponse(
            results_csv=results.to_csv(index=False), calculated_csv=calculated.to_csv(index=False)
        )

    def filterDataset(self, request_iterator, context: grpc.ServicerContext):
        filterfunc = {}
        meta = None

        times = []  # This is an array of chunk execution times for performance stats
        column_types = []

        for filter_msg in request_iterator:
            if filter_msg.HasField("meta"):
                meta = filter_msg.meta
                try:
                    exec(meta.function, None, filterfunc)
                except Exception as e:
                    yield FilterDatasetResponse(code=StatusCode.Failed, error_message=str(e))
                column_types = meta.column_types
                logger.info(f"Filtering dataset with {meta}")

                def filter_wrapper(row):
                    try:
                        return bool(filterfunc["filter_row"](row))
                    except Exception as e:  # noqa # NOSONAR
                        raise ValueError("Failed to execute user defined filtering function") from e

                yield FilterDatasetResponse(code=StatusCode.Ready)
            elif filter_msg.HasField("data"):
                logger.info("Got chunk " + str(filter_msg.idx))
                time_start = time.perf_counter()
                data_as_string = filter_msg.data.content.decode("utf-8")
                data_as_string = meta.headers + "\n" + data_as_string
                # CSV => Dataframe
                data = StringIO(data_as_string)  # Wrap using StringIO to avoid creating file
                df = pd.read_csv(data, keep_default_na=False, na_values=["_GSK_NA_"])
                df = df.astype(column_types)
                # Iterate over rows, applying filter_row func
                try:
                    rows_to_keep = df[df.apply(filter_wrapper, axis=1)].index.array
                except Exception as e:
                    yield FilterDatasetResponse(code=StatusCode.Failed, error_message=str(e))
                time_end = time.perf_counter()
                times.append(time_end - time_start)
                # Send NEXT code
                yield FilterDatasetResponse(
                    code=StatusCode.Next, idx=filter_msg.idx, rows=rows_to_keep
                )

        logger.info(f"Filter dataset finished. Avg chunk time: {sum(times) / len(times)}")
        yield FilterDatasetResponse(code=StatusCode.Ok)

    @staticmethod
    def pandas_df_to_proto_df(df):
        return DataFrame(rows=[DataRow(columns=r.astype(str).to_dict()) for _, r in df.iterrows()])

    @staticmethod
    def pandas_series_to_proto_series(self, series):
        return



import asyncio
import collections
import hashlib
import logging
import os
import sys
from time import sleep

from daemon import DaemonContext
from daemon.daemon import change_working_directory
from daemon.runner import is_pidfile_stale
from lockfile.pidlockfile import PIDLockFile

from giskard.ml_worker.ml_worker import start_ml_worker
from giskard.path_utils import run_dir
from giskard.settings import settings

logger = logging.getLogger(__name__)


def remove_stale_pid_file(pid_file):
    if is_pidfile_stale(pid_file):
        logger.debug("Stale PID file found, removing it")
        pid_file.break_lock()


def create_pid_file_path(is_server, host, port):
    key = f"{sys.executable}"
    if not is_server:
        key += f"{host}{port}"
    hash_value = hashlib.sha1(key.encode()).hexdigest()
    return run_dir / f"ml-worker-{hash_value}.pid"


def run_daemon(is_server, host, port):
    log_path = get_log_path()
    logger.info(f"Writing logs to {log_path}")
    pid_file = PIDLockFile(create_pid_file_path(is_server, host, port))

    with DaemonContext(pidfile=pid_file, stdout=open(log_path, "w+t")):
        workdir = settings.home_dir / "tmp" / f"daemon-run-{os.getpid()}"
        workdir.mkdir(exist_ok=True, parents=True)
        change_working_directory(workdir)

        logger.info(f"Daemon PID: {os.getpid()}")
        asyncio.get_event_loop().run_until_complete(start_ml_worker(is_server, host, port))


def get_log_path():
    return run_dir / "ml-worker.log"


def tail(filename, n=100):
    """Return the last n lines of a file"""
    return collections.deque(open(filename), n)


def follow_file(filename):
    if not os.path.exists(filename):
        print(f"{filename} does not exists")
        return

    wait = 1
    with open(filename) as fp:
        exit_pooling = False
        skip_existing = True
        while not exit_pooling:
            line = fp.readline()

            if not line:
                skip_existing = False
                sleep(wait)
            elif not skip_existing:
                print(line, end="")



import pandas as pd
import numpy as np
from typing import Callable

from giskard.ml_worker.core.giskard_dataset import GiskardDataset
from giskard.ml_worker.core.model import GiskardModel
from giskard.ml_worker.generated.ml_worker_pb2 import SingleTestResult, TestMessage, TestMessageType
from giskard.ml_worker.testing.abstract_test_collection import AbstractTestCollection


class StatisticalTests(AbstractTestCollection):
    def test_right_label(
            self,
            actual_slice: GiskardDataset,
            model: GiskardModel,
            classification_label: str,
            threshold=0.5,
    ) -> SingleTestResult:
        """
        Summary: Test if the model returns the right classification label for a slice

        Description: The test is passed when the percentage of rows returning the right
        classification label is higher than the threshold in a given slice

        Example: For a credit scoring model, the test is passed when more than 50%
        of people with high-salaries are classified as “non default”

        Args:
           actual_slice(GiskardDataset):
              Slice of the  actual dataset
          model(GiskardModel):
              Model used to compute the test
          classification_label(str):
              Classification label you want to test
          threshold(float):
              Threshold for the percentage of passed rows

        Returns:
          actual_slices_size:
              Length of actual_slice tested
          metrics:
              The ratio of rows with the right classification label over the total of rows in the slice
          passed:
              TRUE if passed_ratio > threshold
        """
        actual_slice.df.reset_index(drop=True, inplace=True)
        prediction_results = model.run_predict(actual_slice).prediction
        assert (
                classification_label in model.classification_labels
        ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'

        passed_idx = actual_slice.df.loc[prediction_results == classification_label].index.values

        passed_ratio = len(passed_idx) / len(actual_slice)
        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(actual_slice)],
                metric=passed_ratio,
                passed=passed_ratio > threshold,
            )
        )

    def test_output_in_range(
            self,
            actual_slice: GiskardDataset,
            model: GiskardModel,
            classification_label=None,
            min_range: float = 0.3,
            max_range: float = 0.7,
            threshold=0.5,
    ) -> SingleTestResult:
        """
        Summary: Test if the model output belongs to the right range for a slice

        Description: - The test is passed when the ratio of rows in the right range inside the
        slice is higher than the threshold.

         For classification: Test if the predicted probability for a given classification label
         belongs to the right range for a dataset slice

        For regression : Test if the predicted output belongs to the right range for a dataset slice

        Example :
        For Classification: For a credit scoring model, the test is passed when more than 50% of
        people with high wage have a probability of defaulting between 0 and 0.1

        For Regression : The predicted Sale Price of a house in the city falls in a particular range
        Args:
            actual_slice(GiskardDataset):
                Slice of the actual dataset
            model(GiskardModel):
                Model used to compute the test
            classification_label(str):
                Optional. Classification label you want to test
            min_range(float):
                Minimum probability of occurrence of classification label
            max_range(float):
                Maximum probability of occurrence of classification label
            threshold(float):
                Threshold for the percentage of passed rows

        Returns:
            actual_slices_size:
                Length of actual_slice tested
            metrics:
                The proportion of rows in the right range inside the slice
            passed:
                TRUE if metric > threshold
        """
        results_df = pd.DataFrame()
        actual_slice.df.reset_index(drop=True, inplace=True)

        prediction_results = model.run_predict(actual_slice)

        if model.model_type == "regression":
            results_df["output"] = prediction_results.raw_prediction

        elif model.model_type == "classification":
            assert (
                    classification_label in model.classification_labels
            ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'
            results_df["output"] = prediction_results.all_predictions[classification_label]

        else:
            raise ValueError(f"Prediction task is not supported: {model.model_type}")

        passed_idx = actual_slice.df.loc[
            (results_df["output"] <= max_range) & (results_df["output"] >= min_range)
            ].index.values

        passed_ratio = len(passed_idx) / len(actual_slice)

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(actual_slice)],
                metric=passed_ratio,
                passed=passed_ratio >= threshold,
            )
        )

    def test_disparate_impact(
            self,
            gsk_dataset: GiskardDataset,
            protected_slice: Callable[[pd.DataFrame], pd.DataFrame],
            unprotected_slice: Callable[[pd.DataFrame], pd.DataFrame],
            model: GiskardModel,
            positive_outcome,
            min_threshold=0.8,
            max_threshold=1.25,
    ) -> SingleTestResult:

        """
        Summary: Tests if the model is biased more towards an unprotected slice of the dataset over a protected slice.
        Note that this test reflects only a possible bias in the model while being agnostic to any bias in the dataset
        it trained on. The Disparate Impact (DI) is only valid for classification models and is computed as the ratio
        between the average count of correct predictions for the protected_slice over the unprotected_slice given a
        certain positive_outcome.

        Description: Calculate the Disparate Impact between a protected and unprotected slice of a dataset. Otherwise
        known as the "80 percent" rule, the Disparate Impact determines if a model was having an "adverse impact" on a
        protected (or minority in some cases) group.

        Example: The rule was originally based on the rates at which job applicants were hired. For example, if XYZ
        Company hired 50 percent of the men applying for work in a predominantly male occupation while hiring only 20
        percent of the female applicants, one could look at the ratio of those two hiring rates to judge whether there
        might be a discrimination problem. The ratio of 20:50 means that the rate of hiring for female applicants is
        only 40 percent of the rate of hiring for male applicants. That is, 20 divided by 50 equals
        0.40, which is equivalent to 40 percent. Clearly, 40 percent is well below the 80 percent that was arbitrarily
        set as an acceptable difference in hiring rates. Therefore, in this example, XYZ Company could have been called
        upon to prove that there was a legitimate reason for hiring men at a rate so much higher than the rate of hiring
        women.

        Args:
              gsk_dataset(GiskardDataset):
                  Dataset used to compute the test
              protected_slice(Callable):
                  Slice that defines the protected group from the full dataset given
              unprotected_slice(Callable):
                  Slice that defines the unprotected group from the full dataset given
              model(GiskardModel):
                  Model used to compute the test
              positive_outcome(str or float):
                  The target value that is considered a positive outcome in the dataset
              min_threshold(float):
                  Threshold below which the DI test is considered to fail, by default 0.8
              max_threshold(float):
                  Threshold above which the DI test is considered to fail, by default 1.25

        Returns:
              metric:
                  The disparate impact ratio
              passed:
                  TRUE if the disparate impact ratio > min_threshold && disparate impact ratio < max_threshold
        """

        if positive_outcome not in gsk_dataset.df[gsk_dataset.target].values:
            raise ValueError(
                f"The positive outcome chosen {positive_outcome} is not part of the dataset columns {gsk_dataset.columns}."
            )

        gsk_dataset.df.reset_index(drop=True, inplace=True)
        protected_ds = gsk_dataset.slice(protected_slice)
        unprotected_ds = gsk_dataset.slice(unprotected_slice)

        if protected_ds.df.equals(unprotected_ds.df):
            raise ValueError(
                "The protected and unprotected datasets are equal. Please check that you chose different slices."
            )

        positive_idx = list(model.classification_labels).index(positive_outcome)

        protected_predictions = np.squeeze(
            model.run_predict(protected_ds).raw_prediction == positive_idx
        )
        unprotected_predictions = np.squeeze(
            model.run_predict(unprotected_ds).raw_prediction == positive_idx
        )

        protected_proba = np.count_nonzero(protected_predictions) / len(protected_ds.df)
        unprotected_proba = np.count_nonzero(unprotected_predictions) / len(unprotected_ds.df)
        disparate_impact_score = protected_proba / unprotected_proba

        messages = [
            TestMessage(
                type=TestMessageType.INFO,
                text=f"min_threshold = {min_threshold}, max_threshold = {max_threshold}",
            )
        ]

        return self.save_results(
            SingleTestResult(
                metric=disparate_impact_score,
                passed=(disparate_impact_score > min_threshold) * (disparate_impact_score < max_threshold),
                messages=messages,
            )
        )



import logging
from itertools import groupby
from typing import Callable, Dict, List, Any

import eli5
import numpy as np
import pandas as pd
import shap
from eli5.lime import TextExplainer

from giskard.ml_worker.core.giskard_dataset import GiskardDataset
from giskard.ml_worker.core.model import GiskardModel
from giskard.ml_worker.utils.logging import timer

logger = logging.getLogger(__name__)


@timer()
def explain(model: GiskardModel, dataset: GiskardDataset, input_data: Dict):
    def prepare_df(df):
        prepared_df = model.prepare_dataframe(
            GiskardDataset(
                df=df,
                target=dataset.target,
                feature_types=dataset.feature_types,
                column_types=dataset.column_types,
            )
        )
        columns_in_original_order = (
            model.feature_names
            if model.feature_names
            else [c for c in dataset.df.columns if c in prepared_df.columns]
        )
        # Make sure column order is the same as in df
        return prepared_df[columns_in_original_order]

    df = model.prepare_dataframe(dataset)
    feature_names = list(df.columns)

    input_df = prepare_df(pd.DataFrame([input_data]))

    def predict_array(array):
        return model.prediction_function(prepare_df(pd.DataFrame(array, columns=list(df.columns))))

    example = background_example(df, dataset.feature_types)
    kernel = shap.KernelExplainer(predict_array, example)
    shap_values = kernel.shap_values(input_df, silent=True)

    if model.model_type == "regression":
        explanation_chart_data = summary_shap_regression(
            shap_values=shap_values, feature_names=feature_names
        )
    elif model.model_type == "classification":
        explanation_chart_data = summary_shap_classification(
            shap_values=shap_values,
            feature_names=feature_names,
            class_names=model.classification_labels,
        )
    else:
        raise ValueError(f"Prediction task is not supported: {model.model_type}")
    return explanation_chart_data


@timer()
def explain_text(
        model: GiskardModel,
        input_df: pd.DataFrame,
        text_column: str,
        text_document: str,
        n_samples: int,
):
    text_explainer = TextExplainer(random_state=42, n_samples=n_samples)
    prediction_function = text_explanation_prediction_wrapper(
        model.prediction_function, input_df, text_column
    )
    text_explain_attempts = 10
    try:
        for i in range(text_explain_attempts):
            try:
                text_explainer.fit(text_document, prediction_function)
                break
            except ZeroDivisionError:
                logger.warning(f"Failed to fit text explainer {i}")

        text_explainer.show_prediction(target_names=model.classification_labels)
        exp = text_explainer.explain_prediction(target_names=model.classification_labels)
        exp = eli5.formatters.html.prepare_weighted_spans(exp.targets)
        return get_list_words_weights(exp)

    except Exception as e:
        logger.exception(f"Failed to explain text: {text_document}", e)
        raise Exception("Failed to create text explanation") from e


def get_list_words_weights(exp):
    list_words = []
    document = exp[0][0].doc_weighted_spans.document
    for k, g in groupby(document, str.isalnum):
        list_words.append("".join(g))
    list_weights = []
    for target in exp:
        current_weights = []
        t = target[0]
        i = 0
        for word in list_words:
            weight = t.char_weights[i]
            current_weights.append(weight)
            i += len(word)
        list_weights.append(current_weights)
    return (list_words, list_weights)


def background_example(df: pd.DataFrame, input_types: Dict[str, str]) -> pd.DataFrame:
    example = df.mode(dropna=False).head(1)  # si plusieurs modes, on prend le premier
    example.fillna("", inplace=True)
    median = df.median()
    num_columns = [key for key in list(df.columns) if input_types.get(key) == "numeric"]
    for column in num_columns:
        example[column] = median[column]
    return example.astype(df.dtypes)


def summary_shap_classification(
        shap_values: List[np.ndarray],
        feature_names: List[str],
        class_names: List[str],
        max_display: int = 5,
) -> Dict[str, Dict[str, Dict[str, float]]]:
    feature_order = np.argsort(np.sum(np.mean(np.abs(shap_values), axis=1), axis=0))
    feature_order = feature_order[-min(max_display, len(feature_order)):]
    feature_inds = feature_order[:max_display]
    chart_data: Dict[str, Dict[Any, Any]] = {"explanations": {}}
    for i in range(len(shap_values)):
        global_shap_values = np.abs(shap_values[i]).mean(0)
        chart_data["explanations"][class_names[i]] = {
            feature_names[feature_ind]: global_shap_values[feature_ind]
            for feature_ind in feature_inds
        }
    return chart_data


def summary_shap_regression(
        shap_values: np.ndarray, feature_names: List[str], max_display: int = 5
) -> Dict[str, Dict[str, Dict[str, float]]]:
    max_display = min(max_display, shap_values.shape[1])
    feature_order = np.argsort(np.mean(np.abs(shap_values), axis=0))
    feature_order = feature_order[-max_display:]
    feature_inds = feature_order[:max_display]
    global_shap_values = np.abs(shap_values).mean(0)

    chart_data = {
        "explanations": {
            "default": {
                feature_names[feature_ind]: global_shap_values[feature_ind]
                for feature_ind in feature_inds
            }
        }
    }
    return chart_data


def text_explanation_prediction_wrapper(
        prediction_function: Callable, input_example: pd.DataFrame, text_column: str
) -> Callable:
    def text_predict(text_documents: List[str]):
        num_documents = len(text_documents)

        df_with_text_documents = input_example.append(
            [input_example] * (num_documents - 1), ignore_index=True
        )
        df_with_text_documents[text_column] = pd.DataFrame(text_documents)
        return prediction_function(df_with_text_documents)

    return text_predict



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import collections.abc
import google.protobuf.empty_pb2
import grpc
import ml_worker_pb2

class MLWorkerStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    getInfo: grpc.UnaryUnaryMultiCallable[
        ml_worker_pb2.MLWorkerInfoRequest,
        ml_worker_pb2.MLWorkerInfo,
    ]
    runTest: grpc.UnaryUnaryMultiCallable[
        ml_worker_pb2.RunTestRequest,
        ml_worker_pb2.TestResultMessage,
    ]
    runAdHocTest: grpc.UnaryUnaryMultiCallable[
        ml_worker_pb2.RunAdHocTestRequest,
        ml_worker_pb2.TestResultMessage,
    ]
    runModel: grpc.UnaryUnaryMultiCallable[
        ml_worker_pb2.RunModelRequest,
        ml_worker_pb2.RunModelResponse,
    ]
    runModelForDataFrame: grpc.UnaryUnaryMultiCallable[
        ml_worker_pb2.RunModelForDataFrameRequest,
        ml_worker_pb2.RunModelForDataFrameResponse,
    ]
    explain: grpc.UnaryUnaryMultiCallable[
        ml_worker_pb2.ExplainRequest,
        ml_worker_pb2.ExplainResponse,
    ]
    explainText: grpc.UnaryUnaryMultiCallable[
        ml_worker_pb2.ExplainTextRequest,
        ml_worker_pb2.ExplainTextResponse,
    ]
    echo: grpc.UnaryUnaryMultiCallable[
        ml_worker_pb2.EchoMsg,
        ml_worker_pb2.EchoMsg,
    ]
    upload: grpc.StreamStreamMultiCallable[
        ml_worker_pb2.FileUploadRequest,
        ml_worker_pb2.UploadStatus,
    ]
    getTestRegistry: grpc.UnaryUnaryMultiCallable[
        google.protobuf.empty_pb2.Empty,
        ml_worker_pb2.TestRegistryResponse,
    ]
    filterDataset: grpc.StreamStreamMultiCallable[
        ml_worker_pb2.FilterDatasetRequest,
        ml_worker_pb2.FilterDatasetResponse,
    ]

class MLWorkerServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def getInfo(
        self,
        request: ml_worker_pb2.MLWorkerInfoRequest,
        context: grpc.ServicerContext,
    ) -> ml_worker_pb2.MLWorkerInfo: ...
    @abc.abstractmethod
    def runTest(
        self,
        request: ml_worker_pb2.RunTestRequest,
        context: grpc.ServicerContext,
    ) -> ml_worker_pb2.TestResultMessage: ...
    @abc.abstractmethod
    def runAdHocTest(
        self,
        request: ml_worker_pb2.RunAdHocTestRequest,
        context: grpc.ServicerContext,
    ) -> ml_worker_pb2.TestResultMessage: ...
    @abc.abstractmethod
    def runModel(
        self,
        request: ml_worker_pb2.RunModelRequest,
        context: grpc.ServicerContext,
    ) -> ml_worker_pb2.RunModelResponse: ...
    @abc.abstractmethod
    def runModelForDataFrame(
        self,
        request: ml_worker_pb2.RunModelForDataFrameRequest,
        context: grpc.ServicerContext,
    ) -> ml_worker_pb2.RunModelForDataFrameResponse: ...
    @abc.abstractmethod
    def explain(
        self,
        request: ml_worker_pb2.ExplainRequest,
        context: grpc.ServicerContext,
    ) -> ml_worker_pb2.ExplainResponse: ...
    @abc.abstractmethod
    def explainText(
        self,
        request: ml_worker_pb2.ExplainTextRequest,
        context: grpc.ServicerContext,
    ) -> ml_worker_pb2.ExplainTextResponse: ...
    @abc.abstractmethod
    def echo(
        self,
        request: ml_worker_pb2.EchoMsg,
        context: grpc.ServicerContext,
    ) -> ml_worker_pb2.EchoMsg: ...
    @abc.abstractmethod
    def upload(
        self,
        request_iterator: collections.abc.Iterator[ml_worker_pb2.FileUploadRequest],
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[ml_worker_pb2.UploadStatus]: ...
    @abc.abstractmethod
    def getTestRegistry(
        self,
        request: google.protobuf.empty_pb2.Empty,
        context: grpc.ServicerContext,
    ) -> ml_worker_pb2.TestRegistryResponse: ...
    @abc.abstractmethod
    def filterDataset(
        self,
        request_iterator: collections.abc.Iterator[ml_worker_pb2.FilterDatasetRequest],
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[ml_worker_pb2.FilterDatasetResponse]: ...

def add_MLWorkerServicer_to_server(servicer: MLWorkerServicer, server: grpc.Server) -> None: ...



import pandas as pd

from giskard.ml_worker.core.giskard_dataset import GiskardDataset
from giskard.ml_worker.core.model import GiskardModel
from giskard.ml_worker.generated.ml_worker_pb2 import SingleTestResult, TestMessage, TestMessageType
from giskard.ml_worker.testing.abstract_test_collection import AbstractTestCollection
from giskard.ml_worker.testing.utils import apply_perturbation_inplace
from giskard.ml_worker.utils.logging import timer
from giskard.ml_worker.testing.stat_utils import equivalence_t_test, paired_t_test
from giskard.ml_worker.testing.stat_utils import equivalence_wilcoxon, paired_wilcoxon
from giskard.ml_worker.testing.utils import Direction


class MetamorphicTests(AbstractTestCollection):
    @staticmethod
    def _predict_numeric_result(
        ds: GiskardDataset, model: GiskardModel, output_proba=True, classification_label=None
    ):
        if model.model_type == "regression" or not output_proba:
            return model.run_predict(ds).raw_prediction
        elif model.model_type == "classification" and classification_label is not None:
            return model.run_predict(ds).all_predictions[classification_label].values
        elif model.model_type == "classification":
            return model.run_predict(ds).probabilities

    @staticmethod
    def _prediction_ratio(prediction, perturbed_prediction):
        return (
            abs(perturbed_prediction - prediction) / prediction
            if prediction != 0
            else abs(perturbed_prediction)
        )

    @staticmethod
    def _perturb_and_predict(
        ds: GiskardDataset,
        model: GiskardModel,
        perturbation_dict,
        output_proba=True,
        classification_label=None,
    ):
        results_df = pd.DataFrame()
        results_df["prediction"] = MetamorphicTests._predict_numeric_result(
            ds, model, output_proba, classification_label
        )
        modified_rows = apply_perturbation_inplace(ds.df, perturbation_dict)
        if len(modified_rows):
            ds.df = ds.df.iloc[modified_rows]
            results_df = results_df.iloc[modified_rows]
            results_df["perturbed_prediction"] = MetamorphicTests._predict_numeric_result(
                ds, model, output_proba, classification_label
            )
        else:
            results_df["perturbed_prediction"] = results_df["prediction"]
        return results_df, len(modified_rows)

    @timer("Compare and predict the data")
    def _compare_prediction(self, results_df, prediction_task, direction, output_sensitivity=None):
        if direction == Direction.Invariant:
            if prediction_task == "classification":
                passed_idx = results_df.loc[
                    results_df["prediction"] == results_df["perturbed_prediction"]
                ].index.values

            elif prediction_task == "regression":
                results_df["predict_difference_ratio"] = results_df.apply(
                    lambda x: self._prediction_ratio(x["prediction"], x["perturbed_prediction"]),
                    axis=1,
                )
                passed_idx = results_df.loc[
                    results_df["predict_difference_ratio"] < output_sensitivity
                ].index.values

        elif direction == Direction.Increasing:
            passed_idx = results_df.loc[
                results_df["prediction"] < results_df["perturbed_prediction"]
            ].index.values

        elif direction == Direction.Decreasing:
            passed_idx = results_df.loc[
                results_df["prediction"] > results_df["perturbed_prediction"]
            ].index.values

        failed_idx = results_df.loc[~results_df.index.isin(passed_idx)].index.values
        return passed_idx, failed_idx

    def _compare_probabilities_t_test(
        self, result_df, direction, window_size=0.1, critical_quantile=0.05
    ):

        if direction == Direction.Invariant:
            p_value = equivalence_t_test(
                result_df["prediction"],
                result_df["perturbed_prediction"],
                window_size=window_size,
                critical_quantile=critical_quantile,
            )[1]

        elif direction == Direction.Increasing:
            p_value = paired_t_test(
                result_df["prediction"],
                result_df["perturbed_prediction"],
                alternative="less",
                critical_quantile=critical_quantile,
            )[1]

        elif direction == Direction.Decreasing:
            p_value = paired_t_test(
                result_df["prediction"],
                result_df["perturbed_prediction"],
                alternative="greater",
                critical_quantile=critical_quantile,
            )[1]

        return p_value

    def _compare_probabilities_wilcoxon(
        self, result_df, direction, window_size=0.2, critical_quantile=0.05
    ):

        if direction == Direction.Invariant:
            p_value = equivalence_wilcoxon(
                result_df["prediction"],
                result_df["perturbed_prediction"],
                window_size=window_size,
                critical_quantile=critical_quantile,
            )[1]

        elif direction == Direction.Increasing:
            p_value = paired_wilcoxon(
                result_df["prediction"],
                result_df["perturbed_prediction"],
                alternative="less",
                critical_quantile=critical_quantile,
            )[1]

        elif direction == Direction.Decreasing:
            p_value = paired_wilcoxon(
                result_df["prediction"],
                result_df["perturbed_prediction"],
                alternative="greater",
                critical_quantile=critical_quantile,
            )[1]

        return p_value

    def _test_metamorphic(
        self,
        direction: Direction,
        actual_slice: GiskardDataset,
        model,
        perturbation_dict,
        threshold: float,
        classification_label=None,
        output_sensitivity=None,
        output_proba=True,
    ) -> SingleTestResult:
        actual_slice.df.reset_index(drop=True, inplace=True)

        results_df, modified_rows_count = self._perturb_and_predict(
            actual_slice,
            model,
            perturbation_dict,
            classification_label=classification_label,
            output_proba=output_proba,
        )

        passed_idx, failed_idx = self._compare_prediction(
            results_df, model.model_type, direction, output_sensitivity
        )
        passed_ratio = len(passed_idx) / modified_rows_count if modified_rows_count != 0 else 1

        messages = [
            TestMessage(
                type=TestMessageType.INFO, text=f"{modified_rows_count} rows were perturbed"
            )
        ]

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(actual_slice)],
                metric=passed_ratio,
                passed=passed_ratio > threshold,
                messages=messages,
            )
        )

    def test_metamorphic_invariance(
        self, df: GiskardDataset, model, perturbation_dict, threshold=0.5, output_sensitivity=None
    ) -> SingleTestResult:
        """
        Summary: Tests if the model prediction is invariant when the feature values are perturbed

        Description: -
        For classification: Test if the predicted classification label remains the same after
        feature values perturbation.
        For regression: Check whether the predicted output remains the same at the output_sensibility
        level after feature values perturbation.

        The test is passed when the ratio of invariant rows is higher than the threshold

        Example : The test is passed when, after switching gender from male to female,
        more than 50%(threshold 0.5) of males have unchanged outputs

        Args:
            df(GiskardDataset):
              Dataset used to compute the test
            model(GiskardModel):
              Model used to compute the test
            perturbation_dict(dict):
              Dictionary of the perturbations. It provides the perturbed features as key
              and a perturbation lambda function as value
            window_size(float):
              Threshold of the ratio of invariant rows
            output_sensitivity(float):
                Optional. The threshold for ratio between the difference between perturbed prediction and actual prediction over
                the actual prediction for a regression model. We consider there is a prediction difference for
                regression if the ratio is above the output_sensitivity of 0.1

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            message:
              Test result message
            metric:
              The ratio of unchanged rows over the perturbed rows
            passed:
              TRUE if metric > threshold
        """

        return self._test_metamorphic(
            direction=Direction.Invariant,
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            threshold=threshold,
            output_sensitivity=output_sensitivity,
            output_proba=False,
        )

    def test_metamorphic_increasing(
        self, df: GiskardDataset, model, perturbation_dict, threshold=0.5, classification_label=None
    ):
        """
        Summary: Tests if the model probability increases when the feature values are perturbed

        Description: -
        - For classification: Test if the model probability of a given classification_label is
        increasing after feature values perturbation.

        - For regression: Test if the model prediction is increasing after feature values perturbation.

        The test is passed when the percentage of rows that are increasing is higher than the threshold

        Example : For a credit scoring model, the test is passed when a decrease of wage by 10%,
         default probability is increasing for more than 50% of people in the dataset

        Args:
            df(GiskardDataset):
              Dataset used to compute the test
            model(GiskardModel):
              Model used to compute the test
            perturbation_dict(dict):
              Dictionary of the perturbations. It provides the perturbed features as key
              and a perturbation lambda function as value
            window_size(float):
              Threshold of the ratio of increasing rows
            classification_label(str):
              Optional.One specific label value from the target column

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            message:
              Test result message
            metric:
              The ratio of increasing rows over the perturbed rows
            passed:
              TRUE if metric > threshold
        """
        assert (
            model.model_type != "classification"
            or str(classification_label) in model.classification_labels
        ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'

        return self._test_metamorphic(
            direction=Direction.Increasing,
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            classification_label=classification_label,
            threshold=threshold,
        )

    def test_metamorphic_decreasing(
        self, df: GiskardDataset, model, perturbation_dict, threshold=0.5, classification_label=None
    ):
        """
        Summary: Tests if the model probability decreases when the feature values are perturbed

        Description: -
        - For classification: Test if the model probability of a given classification_label is
        decreasing after feature values perturbation.

        - For regression: Test if the model prediction is decreasing after feature values perturbation.

        The test is passed when the percentage of rows that are decreasing is higher than the threshold

        Example : For a credit scoring model, the test is passed when an increase of wage by 10%,
         default probability is decreasing for more than 50% of people in the dataset

        Args:
            df(GiskardDataset):
              Dataset used to compute the test
            model(GiskardModel):
              Model used to compute the test
            perturbation_dict(dict):
              Dictionary of the perturbations. It provides the perturbed features as key
              and a perturbation lambda function as value
            threshold(float):
              Threshold of the ratio of decreasing rows
            classification_label(str):
              Optional. One specific label value from the target column

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            message:
              Test result message
            metric:
              The ratio of decreasing rows over the perturbed rows
            passed:
              TRUE if metric > threshold
        """

        assert (
            model.model_type != "classification"
            or classification_label in model.classification_labels
        ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'

        return self._test_metamorphic(
            direction=Direction.Decreasing,
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            classification_label=classification_label,
            threshold=threshold,
        )

    def _test_metamorphic_t_test(
        self,
        direction: Direction,
        actual_slice: GiskardDataset,
        model,
        perturbation_dict,
        window_size: float,
        critical_quantile: float,
        classification_label=None,
        output_proba=True,
    ) -> SingleTestResult:

        actual_slice.df.reset_index(drop=True, inplace=True)

        result_df, modified_rows_count = self._perturb_and_predict(
            actual_slice,
            model,
            perturbation_dict,
            output_proba=output_proba,
            classification_label=classification_label,
        )

        p_value = self._compare_probabilities_t_test(
            result_df, direction, window_size, critical_quantile
        )

        messages = [
            TestMessage(
                type=TestMessageType.INFO, text=f"{modified_rows_count} rows were perturbed"
            )
        ]

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(actual_slice)],
                metric=p_value,
                passed=p_value < critical_quantile,
                messages=messages,
            )
        )

    def test_metamorphic_decreasing_t_test(
        self,
        df: GiskardDataset,
        model,
        perturbation_dict,
        critical_quantile=0.05,
        classification_label=None,
    ):
        """
        Summary: Tests if the model probability decreases when the feature values are perturbed

        Description: Calculate the t-test on TWO RELATED samples. Sample (A) is the original probability predictions
        while sample (B) is the probabilities after perturbation of one or more of the features.
        This test computes the decreasing test to study if mean(B) < mean(A)
        The test is passed when the p-value of the t-test between (A) and (B) is below the critical quantile

        Example: For a credit scoring model, the test is passed when a decrease of wage by 10%,
                 causes a statistically significant probability decrease.

        Args:
            df(GiskardDataset):
                Dataset used to compute the test
            model(GiskardModel):
                Model used to compute the test
            perturbation_dict(dict):
                Dictionary of the perturbations. It provides the perturbed features as key
                and a perturbation lambda function as value
            critical_quantile(float):
                Critical quantile above which the null hypothesis cannot be rejected

        Returns:
            actual_slices_size:
                Length of actual_slice tested
            message:
                Test result message
            metric:
                The t-test in terms of p-value between unchanged rows over the perturbed rows
            passed:
                TRUE if the p-value of the t-test between (A) and (B) is below the critical value
        """

        assert (
            model.model_type != "classification"
            or str(classification_label) in model.classification_labels
        ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'

        return self._test_metamorphic_t_test(
            direction=Direction.Decreasing,
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            classification_label=classification_label,
            window_size=float("nan"),
            critical_quantile=critical_quantile,
        )

    def test_metamorphic_increasing_t_test(
        self,
        df: GiskardDataset,
        model,
        perturbation_dict,
        critical_quantile=0.05,
        classification_label=None,
    ):
        """
        Summary: Tests if the model probability increases when the feature values are perturbed

        Description: Calculate the t-test on TWO RELATED samples. Sample (A) is the original probability predictions
        while sample (B) is the probabilities after perturbation of one or more of the features.
        This test computes the increasing test to study if mean(A) < mean(B)
        The test is passed when the p-value of the t-test between (A) and (B) is below the critical quantile

        Example: For a credit scoring model, the test is passed when a decrease of wage by 10%,
                 causes a statistically significant probability increase.

        Args:
            df(GiskardDataset):
                Dataset used to compute the test
            model(GiskardModel):
                Model used to compute the test
            perturbation_dict(dict):
                Dictionary of the perturbations. It provides the perturbed features as key
                and a perturbation lambda function as value
            critical_quantile(float):
                Critical quantile above which the null hypothesis cannot be rejected

        Returns:
            actual_slices_size:
                Length of actual_slice tested
            message:
                Test result message
            metric:
                The t-test in terms of p-value between unchanged rows over the perturbed rows
            passed:
                TRUE if the p-value of the t-test between (A) and (B) is below the critical value
        """

        assert (
            model.model_type != "classification"
            or str(classification_label) in model.classification_labels
        ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'

        return self._test_metamorphic_t_test(
            direction=Direction.Increasing,
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            classification_label=classification_label,
            window_size=float("nan"),
            critical_quantile=critical_quantile,
        )

    def test_metamorphic_invariance_t_test(
        self,
        df: GiskardDataset,
        model,
        perturbation_dict,
        window_size: float,
        critical_quantile: float,
    ) -> SingleTestResult:
        """
        Summary: Tests if the model predictions are statistically invariant when the feature values are perturbed.

        Description: Calculate the t-test on TWO RELATED samples. Sample (A) is the original probability predictions
        while sample (B) is the probabilities after perturbation of one or more of the features.
        This test computes the equivalence test to show that mean(B) - window_size/2 < mean(A) < mean(B) + window_size/2
        The test is passed when the following tests pass:
          - the p-value of the t-test between (A) and (B)+window_size/2 is below the critical quantile
          - the p-value of the t-test between (B)-window_size/2 and (A) is below the critical quantile

        Example: The test is passed when, after switching gender from male to female,
        the probability distributions remains statistically invariant. In other words, the test is passed if the mean of the
        perturbed sample is statistically within a window determined by the user.

        Args:
              df(GiskardDataset):
                  Dataset used to compute the test
              model(GiskardModel):
                  Model used to compute the test
              perturbation_dict(dict):
                  Dictionary of the perturbations. It provides the perturbed features as key
                  and a perturbation lambda function as value
              window_size(float):
                  Probability window in which the mean of the perturbed sample can be in
              critical_quantile(float):
                  Critical quantile above which the null hypothesis cannot be rejected

        Returns:
              actual_slices_size:
                  Length of actual_slice tested
              message:
                  Test result message
              metric:
                  The t-test in terms of p-value between unchanged rows over the perturbed rows
              passed:
                  TRUE if the p-value of the t-test between (A) and (B)+window_size/2 < critical_quantile && the p-value of the t-test between (B)-window_size/2 and (A) < critical_quantile
        """

        return self._test_metamorphic_t_test(
            direction=Direction.Invariant,
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            window_size=window_size,
            critical_quantile=critical_quantile,
        )

    def _test_metamorphic_wilcoxon(
        self,
        direction: Direction,
        actual_slice: GiskardDataset,
        model,
        perturbation_dict,
        window_size: float,
        critical_quantile: float,
        classification_label=None,
        output_proba=True,
    ) -> SingleTestResult:

        actual_slice.df.reset_index(drop=True, inplace=True)

        result_df, modified_rows_count = self._perturb_and_predict(
            actual_slice,
            model,
            perturbation_dict,
            output_proba=output_proba,
            classification_label=classification_label,
        )

        p_value = self._compare_probabilities_wilcoxon(
            result_df, direction, window_size, critical_quantile
        )

        messages = [
            TestMessage(
                type=TestMessageType.INFO, text=f"{modified_rows_count} rows were perturbed"
            )
        ]

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(actual_slice)],
                metric=p_value,
                passed=p_value < critical_quantile,
                messages=messages,
            )
        )

    def test_metamorphic_decreasing_wilcoxon(
        self,
        df: GiskardDataset,
        model,
        perturbation_dict,
        critical_quantile=0.05,
        classification_label=None,
    ):
        """
        Summary: Tests if the model probability decreases when the feature values are perturbed

        Description: Calculate the Wilcoxon signed-rank test on TWO RELATED samples. Sample (A) is the original probability predictions
        while sample (B) is the probabilities after perturbation of one or more of the features.
        This test computes the decreasing test to study if mean(B) < mean(A)
        The test is passed when the p-value of the Wilcoxon signed-rank test between (A) and (B) is below the critical quantile

        Example: For a credit scoring model, the test is passed when a decrease of wage by 10%,
                 causes a statistically significant probability decrease.

        Args:
            df(GiskardDataset):
                Dataset used to compute the test
            model(GiskardModel):
                Model used to compute the test
            perturbation_dict(dict):
                Dictionary of the perturbations. It provides the perturbed features as key
                and a perturbation lambda function as value
            critical_quantile(float):
                Critical quantile above which the null hypothesis cannot be rejected

        Returns:
            actual_slices_size:
                Length of actual_slice tested
            message:
                Test result message
            metric:
                The Wilcoxon signed-rank test in terms of p-value between unchanged rows over the perturbed rows
            passed:
                TRUE if the p-value of the Wilcoxon signed-rank test between (A) and (B) is below the critical value
        """

        assert (
            model.model_type != "classification"
            or str(classification_label) in model.classification_labels
        ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'

        return self._test_metamorphic_wilcoxon(
            direction=Direction.Decreasing,
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            classification_label=classification_label,
            window_size=float("nan"),
            critical_quantile=critical_quantile,
        )

    def test_metamorphic_increasing_wilcoxon(
        self,
        df: GiskardDataset,
        model,
        perturbation_dict,
        critical_quantile=0.05,
        classification_label=None,
    ):
        """
        Summary: Tests if the model probability increases when the feature values are perturbed

        Description: Calculate the Wilcoxon signed-rank test on TWO RELATED samples. Sample (A) is the original probability predictions
        while sample (B) is the probabilities after perturbation of one or more of the features.
        This test computes the increasing test to study if mean(A) < mean(B)
        The test is passed when the p-value of the Wilcoxon signed-rank test between (A) and (B) is below the critical quantile

        Example: For a credit scoring model, the test is passed when a decrease of wage by 10%,
                 causes a statistically significant probability increase.

        Args:
            df(GiskardDataset):
                Dataset used to compute the test
            model(GiskardModel):
                Model used to compute the test
            perturbation_dict(dict):
                Dictionary of the perturbations. It provides the perturbed features as key
                and a perturbation lambda function as value
            critical_quantile(float):
                Critical quantile above which the null hypothesis cannot be rejected

        Returns:
            actual_slices_size:
                Length of actual_slice tested
            message:
                Test result message
            metric:
                The Wilcoxon signed-rank test in terms of p-value between unchanged rows over the perturbed rows
            passed:
                TRUE if the p-value of the Wilcoxon signed-rank test between (A) and (B) is below the critical value
        """

        assert (
            model.model_type != "classification"
            or str(classification_label) in model.classification_labels
        ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'

        return self._test_metamorphic_wilcoxon(
            direction=Direction.Increasing,
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            classification_label=classification_label,
            window_size=float("nan"),
            critical_quantile=critical_quantile,
        )

    def test_metamorphic_invariance_wilcoxon(
        self, df: GiskardDataset, model, perturbation_dict, window_size=0.2, critical_quantile=0.05
    ) -> SingleTestResult:
        """
        Summary: Tests if the model predictions are statistically invariant when the feature values are perturbed.

        Description: Calculate the Wilcoxon signed-rank test on TWO RELATED samples. Sample (A) is the original probability predictions
        while sample (B) is the probabilities after perturbation of one or more of the features.
        This test computes the equivalence test to show that mean(B) - window_size/2 < mean(A) < mean(B) + window_size/2
        The test is passed when the following tests pass:
        - the p-value of the t-test between (A) and (B)+window_size/2 is below the critical quantile
        - the p-value of the t-test between (B)-window_size/2 and (A) is below the critical quantile

        Example: The test is passed when, after switching gender from male to female,
        the probability distributions remains statistically invariant. In other words, the test is passed if the mean of the
        perturbed sample is statistically within a window determined by the user.

        Args:
            df(GiskardDataset):
                Dataset used to compute the test
            model(GiskardModel):
                Model used to compute the test
            perturbation_dict(dict):
                Dictionary of the perturbations. It provides the perturbed features as key
                and a perturbation lambda function as value
            window_size(float):
                Probability window in which the mean of the perturbed sample can be in
            critical_quantile(float):
                Critical quantile above which the null hypothesis cannot be rejected

        Returns:
            actual_slices_size:
                Length of actual_slice tested
            message:
                Test result message
            metric:
                The t-test in terms of p-value between unchanged rows over the perturbed rows
            passed:
                TRUE if the p-value of the Wilcoxon signed-rank test between (A) and (B)+window_size/2 < critical_quantile && the p-value of the t-test between (B)-window_size/2 and (A) < critical_quantile
        """

        return self._test_metamorphic_wilcoxon(
            direction=Direction.Invariant,
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            window_size=window_size,
            critical_quantile=critical_quantile,
        )
